/************************************************************************/
/*
 4R: 
 1. divide display window into 4 views 
	Output buffer size(display window size) : 1280 x 960
	View size : 640 x 480

	------------------------	   
	|		   |           |
	|   View0  |   View1   |
	|          |           |
	------------------------	   
	|		   |           |
	|   View2  |   View3   |
	|          |           |
	------------------------	   

 2. ePTZ location of each view (these values are just for sample illustration)
	Ceiling mount
	view 0 : Pitch 44, Roll 0		(top)
	view 1 : Pitch 44, Roll 90		(right)
	view 2 : Pitch 44, Roll 180		(bottom)
	view 3 : Pitch 44, Roll 270		(left)

	Floor mount
	view 0 : Pitch -44, Roll 0		(bottom)
	view 1 : Pitch -44, Roll 90		(right)
	view 2 : Pitch -44, Roll 180	(top)
	view 3 : Pitch -44, Roll 270	(left)

	Wall mount
	view 0 : Pitch 0, Yaw -44	(left)
	view 1 : Pitch 0, Yaw 44	(right)
	view 2 : Pitch 44, Yaw 0	(top)
	view 3 : Pitch -44, Yaw 0	(bottom)

 3. use mouse event to control ePTZ

*/
/************************************************************************/
#include <iostream>
#include <stdio.h>
#include "fisheyesw.h"
#include "fisheyesw_errordef.h"

#if defined ( _MSC_VER )
#include "SDL.h"
#pragma comment(lib, "SDL2main.lib")
#pragma comment(lib, "SDL2.lib")
#pragma comment(lib, "d_fisheyesw.lib")
#elif defined( __APPLE__ ) && defined( __MACH__ )
#include <SDL2/SDL.h>
#else
#include "SDL.h"
#include <string.h>
#endif

#ifdef main
#undef main
#endif

#define MULTI_VIEW_COUNT 4

typedef struct _tView
{
	HANDLE hFECtx;
	FEOPTION tOption;
} TFEView;

// mouse event
void EventLoop(TFEView* ptView, SDL_Texture* pTexture, SDL_Renderer *pRenderer);
SCODE UpdateOutVPicture(TFEView* ptView, SDL_Texture* pTexture, SDL_Renderer *pRenderer);

int main(int argc, char *argv[])
{
	if (argc < 3)
	{
		printf("\nUsage:\n");
		printf("\tdewarp4R [image] [mount type]\n\n");
		printf("which\n");
		printf("\timage\t\tonly support bitmap\n");
		printf("\tmount type\t0: wall, 1: ceiling, 2: floor\n");
		return -1;
	}

    const int WINDOW_WIDTH = 1280, WINDOW_HEIGHT = 960;
	const int VIEW_WIDTH = 640, VIEW_HEIGHT = 480;

    char *pszimgPath = NULL;
    int mountType = 0;
    
    // SDL instance
    int sdlRet = 0;
    SDL_Texture *pTexture = NULL;
    SDL_Window *pScreen = NULL;
    SDL_Renderer *pRenderer = NULL;
    SDL_Surface* pImage = NULL;
    SDL_Surface* pBitmap = NULL;
    
    pszimgPath = argv[1];
    mountType = atoi(argv[2]);
    
    // SDL renderer
    sdlRet = SDL_Init(SDL_INIT_VIDEO);
    if (sdlRet)
    {
        return sdlRet;
    }
    if (NULL == (pScreen = SDL_CreateWindow("dewarp4R", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, WINDOW_WIDTH, WINDOW_HEIGHT, 0)))
    {
        return -1;
    }
    if (NULL == (pRenderer = SDL_CreateRenderer(pScreen, -1, 0)))
    {
        return -1;
    }
    if (NULL == (pTexture = SDL_CreateTexture(pRenderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_STREAMING, WINDOW_WIDTH, WINDOW_HEIGHT)))
    {
        return -1;
    }
    
    // Load image file
    if (NULL == (pImage = SDL_LoadBMP(pszimgPath)))
    {
        return -1;
    }
    
    // Convert input image pixel format to RGBA
    SDL_PixelFormat format;
    memset(&format, 0, sizeof(SDL_PixelFormat));
    format.BitsPerPixel = 32;
    format.BytesPerPixel = 4;
    format.Rmask = 0xFF000000; format.Rshift = 0;
    format.Gmask = 0x00FF0000; format.Gshift = 8;
    format.Bmask = 0x0000FF00; format.Bshift = 16;
    format.Amask = 0x000000FF; format.Ashift = 32;
    pBitmap = SDL_ConvertSurface(pImage, &format, 0);
    SDL_FreeSurface(pImage);
    
	// Create output buffer
	DWORD dwDstBufferSize = WINDOW_WIDTH * WINDOW_HEIGHT * 4;
	BYTE *pbyDstBuffer = (BYTE*)malloc(dwDstBufferSize * sizeof(BYTE));
	if (pbyDstBuffer == NULL)
	{
		return -1;
	}
	memset(pbyDstBuffer, 0, dwDstBufferSize * sizeof(BYTE));

	// The left-top point of each view ROI 
	FEPOINT tLeftTopPoints[4] = {{0, 0}, {VIEW_WIDTH, 0}, {0, VIEW_HEIGHT}, {VIEW_WIDTH, VIEW_HEIGHT}};

    // Fisheye library instance
	SCODE scRet = FISHEYE_S_OK;
	TFEView tViews[MULTI_VIEW_COUNT];
	memset(&tViews, 0, sizeof(TFEView) * MULTI_VIEW_COUNT);
	
	for (int i = 0; i < MULTI_VIEW_COUNT; i++)
	{
		// 1. Initialize fisheyesw library
		scRet = Fisheye_Initial(&tViews[i].hFECtx, LIBFISHEYE_VERSION);
		if (FISHEYE_S_OK != scRet)
		{
			goto RELEASE_MEMORY;
		}

		// 2. Set fisheye options
		memset(&tViews[i].tOption, 0, sizeof(FEOPTION));
		tViews[i].tOption.Scale = 100;  // default scale is 100 -> 1x zoom factor

		// Set input fisheye image information
		tViews[i].tOption.InVPicture.Width = pBitmap->w;
		tViews[i].tOption.InVPicture.Height = pBitmap->h;
		tViews[i].tOption.InVPicture.Stride = pBitmap->pitch;
		tViews[i].tOption.InVPicture.Format = FE_PIXELFORMAT_RGB32;
		tViews[i].tOption.InVPicture.Buffer = (BYTE*)pBitmap->pixels;

		// Set output image information (The output picture of each view is the same)
		tViews[i].tOption.OutVPicture.Width = WINDOW_WIDTH;
		tViews[i].tOption.OutVPicture.Height = WINDOW_HEIGHT;
		tViews[i].tOption.OutVPicture.Stride = WINDOW_WIDTH * 4;
		tViews[i].tOption.OutVPicture.Format = FE_PIXELFORMAT_RGB32;
		tViews[i].tOption.OutVPicture.Buffer = pbyDstBuffer;

		// Set FOV information
		tViews[i].tOption.FOVCenter.X = tViews[i].tOption.InVPicture.Width >> 1;
		tViews[i].tOption.FOVCenter.Y = tViews[i].tOption.InVPicture.Height >> 1;
		tViews[i].tOption.FOVRadius = tViews[i].tOption.InVPicture.Height >> 1;

		// Set mount type
		switch (mountType)
		{
		case 0:
			tViews[i].tOption.MountType = FE_MOUNT_WALL;
			break;
		case 1:
			tViews[i].tOption.MountType = FE_MOUNT_CEILING;
			break;
		case 2:
			tViews[i].tOption.MountType = FE_MOUNT_FLOOR;
			break;
		default:
			break;
		}

		// Set dewarp type
		tViews[i].tOption.DewarpType = FE_DEWARP_RECTILINEAR;

		// Set display output region which is different from each view 
		tViews[i].tOption.OutRoi.Left = tLeftTopPoints[i].X;
		tViews[i].tOption.OutRoi.Top = tLeftTopPoints[i].Y;
		tViews[i].tOption.OutRoi.Right = tViews[i].tOption.OutRoi.Left + VIEW_WIDTH;
		tViews[i].tOption.OutRoi.Bottom = tViews[i].tOption.OutRoi.Top + VIEW_HEIGHT;

		// Set flags
		tViews[i].tOption.Flags = (FE_OPTION_INIMAGEHEADER | FE_OPTION_INIMAGEBUFFER | FE_OPTION_OUTIMAGEHEADER | FE_OPTION_OUTIMAGEBUFFER |
			FE_OPTION_FOVCENTER | FE_OPTION_FOVRADIUS | FE_OPTION_OUTROI | FE_OPTION_MOUNTTYPE | FE_OPTION_DEWARPTYPE);
	}
   
	// Set default ePTZ location
	switch (mountType)
	{
	case 0:
		//FE_MOUNT_WALL
		tViews[0].tOption.Pitch = 0;
		tViews[0].tOption.Yaw = -44;
		tViews[1].tOption.Pitch = 0;
		tViews[1].tOption.Yaw = 44;
		tViews[2].tOption.Pitch = 44;
		tViews[2].tOption.Yaw = 0;
		tViews[3].tOption.Pitch = -44;	
		tViews[3].tOption.Yaw = 0;
		tViews[0].tOption.Flags = tViews[1].tOption.Flags = tViews[2].tOption.Flags = tViews[3].tOption.Flags |= (FE_OPTION_YAW | FE_OPTION_PITCH);
		break;
	case 1:
		//FE_MOUNT_CEILING
		tViews[0].tOption.Pitch = tViews[1].tOption.Pitch = tViews[2].tOption.Pitch = tViews[3].tOption.Pitch = 44;
		tViews[0].tOption.Roll = 0;
		tViews[1].tOption.Roll = 90;
		tViews[2].tOption.Roll = 180;
		tViews[3].tOption.Roll = 270;
		tViews[0].tOption.Flags = tViews[1].tOption.Flags = tViews[2].tOption.Flags = tViews[3].tOption.Flags |= (FE_OPTION_ROLL | FE_OPTION_PITCH);
		break;
	case 2:
		//FE_MOUNT_FLOOR
		tViews[0].tOption.Pitch = tViews[1].tOption.Pitch = tViews[2].tOption.Pitch = tViews[3].tOption.Pitch = -44;
		tViews[0].tOption.Roll = 0;
		tViews[1].tOption.Roll = 90;
		tViews[2].tOption.Roll = 180;
		tViews[3].tOption.Roll = 270;
		tViews[0].tOption.Flags = tViews[1].tOption.Flags = tViews[2].tOption.Flags = tViews[3].tOption.Flags |= (FE_OPTION_ROLL | FE_OPTION_PITCH);
		break;
	default:
		break;
	}

	// 3. Update options in fisheyesw library and paint
	UpdateOutVPicture(tViews, pTexture, pRenderer);

    // 4. Wait keyboard and mouse event to perform ePTZ control
    EventLoop(tViews, pTexture, pRenderer);
    
RELEASE_MEMORY:
    
    // release buffer memory by user
	for (int i = 0; i < MULTI_VIEW_COUNT; i++)
	{
		if (tViews[i].hFECtx)
		{
			// Release Fisheye library.
			Fisheye_Release(&tViews[i].hFECtx);
			tViews[i].hFECtx = NULL;
		}
	}
	if (pbyDstBuffer)
	{
		free(pbyDstBuffer);
		pbyDstBuffer = NULL;
	}
    // release SDL resource
    if (pBitmap)
    {
        SDL_FreeSurface(pBitmap);
        pBitmap = NULL;
    }
    if (pTexture)
    {
        SDL_DestroyTexture(pTexture);
        pTexture = NULL;
    }
    if (pRenderer)
    {
        SDL_DestroyRenderer(pRenderer);
        pRenderer = NULL;
    }
    if (pScreen)
    {
        SDL_DestroyWindow(pScreen);
        pScreen = NULL;
    }
    
    SDL_Quit();
    
    return 0;
}

void EventLoop(TFEView* ptViews, SDL_Texture* pTexture, SDL_Renderer *pRenderer)
{
    int bExit = 0;
    SDL_Event evt;
    SDL_MouseButtonEvent beginPos = { 0 };
    int bClick = 0;
    int iFocusViewId = 0;
	
	HANDLE hObject = ptViews[0].hFECtx;
	FEOPTION* ptOption = &ptViews[0].tOption;

    while (!bExit)
    {
        SDL_WaitEvent(&evt);

        switch (evt.type)
        {
            case SDL_MOUSEBUTTONDOWN:
                beginPos = evt.button;  // setup the begin position when mouse click

				// determine which view is focused
				for (int i = 0; i < MULTI_VIEW_COUNT; i++)
				{
					if (beginPos.x > ptViews[i].tOption.OutRoi.Left && beginPos.x <= ptViews[i].tOption.OutRoi.Right 
						&& beginPos.y > ptViews[i].tOption.OutRoi.Top && beginPos.y <= ptViews[i].tOption.OutRoi.Bottom)
					{
						hObject = ptViews[i].hFECtx;
						ptOption = &ptViews[i].tOption;
						break;
					}
				}
						
                bClick = 1;             // enable drag image
                break;
            case SDL_MOUSEBUTTONUP:
                bClick = 0;             // disable drag image
                break;
            case SDL_MOUSEMOTION:
                if (bClick)
                {
					// calculate the difference between begin position and current position
					SDL_MouseButtonEvent movePos = evt.button;
					int iDistance_X = (movePos.x - beginPos.x);
					int iDistance_Y = (movePos.y - beginPos.y);

					beginPos = movePos;  // update begin position to current position

					if (FE_DEWARP_RECTILINEAR == ptOption->DewarpType)
					{
						// move right/left
						if (FE_MOUNT_WALL == ptOption->MountType)
						{
							ptOption->Yaw -= iDistance_X;
							ptOption->Flags = FE_OPTION_YAW;
						}
						else
						{
							ptOption->Roll -= iDistance_X;
							ptOption->Flags = FE_OPTION_ROLL;
						}

						// move up/down
						ptOption->Pitch += iDistance_Y;
						ptOption->Flags |= FE_OPTION_PITCH;
					}
                    
                    // set option and repaint
                    UpdateOutVPicture(ptViews, pTexture, pRenderer);
                }
                break;
            case SDL_QUIT:
                bExit = 1;
                break;
            default:
                break;
        }
    }
}

SCODE UpdateOutVPicture(TFEView* ptViews, SDL_Texture* pTexture, SDL_Renderer *pRenderer)
{
    // set fisheye option, use Flags to control which field of FEOPTION needs to be updated.
	for (int i = 0; i < MULTI_VIEW_COUNT; i++)
	{
		HANDLE hObject = ptViews[i].hFECtx;
		FEOPTION* ptOption = &ptViews[i].tOption;

		SCODE scRet = S_OK;
		scRet = Fisheye_SetOption(hObject, ptOption);
		if (S_OK != scRet)
		{
			return scRet;
		}

		// update dewarp result to OutVPicture.Buffer
		scRet = Fisheye_OneFrame(hObject);
		if (S_OK != scRet)
		{
			return scRet;
		}

		// clear flags
		ptViews[i].tOption.Flags = 0;
	}

    // SDL display 4R
    SDL_UpdateTexture(pTexture, NULL, ptViews[0].tOption.OutVPicture.Buffer, ptViews[0].tOption.OutVPicture.Stride);
    SDL_RenderClear(pRenderer);
    SDL_RenderCopy(pRenderer, pTexture, NULL, NULL);
    SDL_RenderPresent(pRenderer);
    
    return S_OK;
}
